﻿using devDept.Eyeshot;
using devDept.Eyeshot.Entities;
using devDept.Geometry.Entities;
using ID2.Manager.Common;
using ID2.Manager.Data.Models;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ID2.Manager.Controls
{
    public partial class Verification : UserControl
    {
        public delegate void DocumentSelected(Documents doc);
        public event DocumentSelected OnDocumentSelected;

        readonly Informations informations = Informations.Instance;
        readonly string IniFilePath = System.IO.Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), 
            Application.ProductName, $"{Application.ProductName}.ini");

        private static Color EqualColor = Color.FromArgb(44, 44, 44);
        private static Color DiffColor = Color.Yellow;
        private static double Tolerance = 0;

        public Verification()
        {
            InitializeComponent();
            this.radGridViewDocument.AutoSizeColumnsMode = Telerik.WinControls.UI.GridViewAutoSizeColumnsMode.Fill;

            this.Load += Verification_Load;
            this.radGridViewDocument.SelectionChanged += RadGridViewDocument_SelectionChanged;
            this.radBrowseEditorAVEVAPIDFolder.ValueChanged += RadBrowseEditorAVEVAPIDFolder_ValueChanged;
            this.radSpinEditorTolerance.ValueChanged += RadSpinEditorTolerance_ValueChanged;
            this.radColorBoxEqualColor.ValueChanged += RadColorBoxEqualColor_ValueChanged;
            this.radColorBoxDiffColor.ValueChanged += RadColorBoxDiffColor_ValueChanged;

            this.designAutoCAD.ActionMode = actionType.SelectVisibleByPickDynamic;
            this.designAutoCAD.Selection.ColorDynamic = Color.FromArgb(80, Color.OrangeRed);
            this.designAutoCAD.Selection.HaloInnerColor = Color.FromArgb(255, Color.OrangeRed);
            this.designAutoCAD.Selection.HaloOuterColor = Color.FromArgb(64, Color.OrangeRed);
            this.designAutoCAD.Selection.HaloWidthPolygons = 4;
            this.designAutoCAD.Selection.HaloWidthWires = 2;
            this.designAVEVA.ActionMode = actionType.SelectVisibleByPickDynamic;

            #region Camera Sync
            this.designAutoCAD.ActiveViewport.Rotate.Enabled = false;
            this.designAVEVA.ActiveViewport.Rotate.Enabled = false;

            this.designAutoCAD.ActiveViewport.ViewCubeIcon.Visible = false;
            this.designAVEVA.ActiveViewport.ViewCubeIcon.Visible = false;

            this.designAutoCAD.AnimateCamera = false;
            this.designAVEVA.AnimateCamera = false;

            this.designAutoCAD.CameraChangedFrequency = 200;
            this.designAVEVA.CameraChangedFrequency = 200;

            this.designAutoCAD.CameraChanged += CameraChanged;
            this.designAVEVA.CameraChanged += CameraChanged;
            #endregion
        }

        /// <summary>
        /// 서로 다른 엔터티의 색상을 설정한다.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RadColorBoxDiffColor_ValueChanged(object sender, EventArgs e)
        {
            Verification.DiffColor = this.radColorBoxDiffColor.Value;
            string color = $"{Verification.DiffColor.R},{Verification.DiffColor.G},{Verification.DiffColor.B}";
            Classes.ID2Helper.IniWriteValue(IniFilePath, "Verification", "DiffColor", color);
        }

        /// <summary>
        /// 동일한 엔터티의 색상을 설정한다.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RadColorBoxEqualColor_ValueChanged(object sender, EventArgs e)
        {
            Verification.EqualColor = this.radColorBoxEqualColor.Value;
            string color = $"{Verification.EqualColor.R},{Verification.EqualColor.G},{Verification.EqualColor.B}";
            Classes.ID2Helper.IniWriteValue(IniFilePath, "Verification", "EqualColor", color);
        }

        /// <summary>
        /// 수정한 Tolerance를 시스템에 반영한다.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RadSpinEditorTolerance_ValueChanged(object sender, EventArgs e)
        {
            string AVEVAPIDFolder = this.radBrowseEditorAVEVAPIDFolder.Value;
            if (System.IO.Directory.Exists(AVEVAPIDFolder))
            {
                double toler = Convert.ToDouble(this.radSpinEditorTolerance.Value);
                Classes.ID2Helper.IniWriteValue(IniFilePath, "Verification", "Tolerance", toler.ToString());
                Verification.Tolerance = toler;
            }
        }

        private void RadBrowseEditorAVEVAPIDFolder_ValueChanged(object sender, EventArgs e)
        {
            string AVEVAPIDFolder = this.radBrowseEditorAVEVAPIDFolder.Value;
            if (System.IO.Directory.Exists(AVEVAPIDFolder))
            {
                Classes.ID2Helper.IniWriteValue(IniFilePath, "Path", "AVEVA P&ID Folder", AVEVAPIDFolder);
            }
        }

        private void Verification_Load(object sender, EventArgs e)
        {
            string AVEVAPIDFolder = Classes.ID2Helper.IniReadValue(IniFilePath, "Path", "AVEVA P&ID Folder");
            if (!string.IsNullOrEmpty(AVEVAPIDFolder)) this.radBrowseEditorAVEVAPIDFolder.Value = AVEVAPIDFolder;

            string Toler = Classes.ID2Helper.IniReadValue(IniFilePath, "Verification", "Tolerance");
            if (!string.IsNullOrEmpty(Toler))
            {
                this.radSpinEditorTolerance.Value = Convert.ToDecimal(Toler);
                Verification.Tolerance = Convert.ToDouble(this.radSpinEditorTolerance.Value);
            }

            string _EqualColor = Classes.ID2Helper.IniReadValue(IniFilePath, "Verification", "EqualColor");
            if (!string.IsNullOrEmpty(_EqualColor))
            {
                var tokens = _EqualColor.Split(',');
                if (tokens.Length == 3)
                {
                    this.radColorBoxEqualColor.Value =
                         Color.FromArgb(Convert.ToInt32(tokens[0]), Convert.ToInt32(tokens[1]), Convert.ToInt32(tokens[2]));
                }
            }

            string _DiffColor = Classes.ID2Helper.IniReadValue(IniFilePath, "Verification", "DiffColor");
            if (!string.IsNullOrEmpty(_DiffColor))
            {
                var tokens = _DiffColor.Split(',');
                if (tokens.Length == 3)
                {
                    this.radColorBoxDiffColor.Value =
                        Color.FromArgb(Convert.ToInt32(tokens[0]), Convert.ToInt32(tokens[1]), Convert.ToInt32(tokens[2]));
                }
            }
        }

        /// <summary>
        /// 엔터티들의 색상을 바꾼다.
        /// </summary>
        /// <param name="design"></param>
        /// <param name="list"></param>
        public void ColorEntities(Design design, EntityList list, Color color, colorMethodType colorMethod=colorMethodType.byEntity, bool ChangeBlkColor=true)
        {
            foreach (Entity ent in list)
            {
                ColorEntity(design, ent, color, colorMethod, ChangeBlkColor);
            }
        }

        /// <summary>
        /// 엔터티의 색상을 변경한다.
        /// </summary>
        /// <param name="design"></param>
        /// <param name="entity"></param>
        /// <param name="color"></param>
        private void ColorEntity(Design design, Entity entity, Color color, colorMethodType colorMethod = colorMethodType.byEntity, 
            bool ChangeBlkColor = true)
        {
            if (entity is BlockReference blkref)
            {
                blkref.Color = color;
                blkref.ColorMethod = colorMethod;

                if (ChangeBlkColor)
                {
                    var blk = design.Blocks.FirstOrDefault(x => x.Name == blkref.BlockName);
                    if (blk != null)
                    {
                        ColorEntities(design, blk.Entities, color, colorMethodType.byParent);
                        foreach (var attr in blkref.Attributes.Values)
                        {
                            attr.Color = color;
                            attr.ColorMethod = colorMethodType.byParent;
                        }
                    }
                }
            }
            else
            {
                entity.Color = color;
                entity.ColorMethod = colorMethod;
            }
        }

        /// <summary>
        /// 도면을 선택했을때 두 도면을 비교한다.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RadGridViewDocument_SelectionChanged(object sender, EventArgs e)
        {
            /// AutoCAD P&ID 파일을 화면에 표시한다.
            void ShowAutoCADFile(string FilePath)
            {
                this.designAutoCAD.Clear();
                if (System.IO.File.Exists(FilePath))
                {
                    devDept.Eyeshot.Translators.ReadAutodesk ra = new devDept.Eyeshot.Translators.ReadAutodesk(FilePath);
                    ra.DoWork();
                    ra.AddToScene(this.designAutoCAD);

                    var AddEntities = new List<Entity>();
                    this.designAutoCAD.Entities.ForEach(x =>
                    {
                        if (x is LinearPath lp)
                        {
                            int count = Convert.ToInt32(lp.Vertices.Length);
                            for (int i = 0; i < count - 1; ++i)
                            {
                                AddEntities.Add(new devDept.Eyeshot.Entities.Line(lp.Vertices[i], lp.Vertices[i + 1])
                                {
                                    LayerName = lp.LayerName,
                                    LineWeight = lp.LineWeight,
                                    LineTypeMethod = colorMethodType.byEntity
                                });
                            }
                        }
                        /*
                        else if(x is BlockReference blkref && blkref.Attributes.Any())
                        {
                            var entities = blkref.Explode(this.designAutoCAD.Blocks).Where(y => y is devDept.Eyeshot.Entities.Text);
                        }
                        */
                    });
                    this.designAutoCAD.Entities.RemoveAll(x => x is LinearPath);
                    this.designAutoCAD.Entities.AddRange(AddEntities);

                    #region 브랜치가 생성되는 부분에서 라인을 분할
                    var queue = this.designAutoCAD.Entities.Where(x => x is Line).ToList();
                    while(queue.Any())
                    {
                        var line1 = queue.First() as Line;
                        var dir1 = line1.Direction;
                        dir1.Normalize();
                        queue.Remove(line1);
                        for(int i = 0;i < queue.Count;++i)
                        {
                            var line2 = queue.ElementAt(i) as Line;
                            var dir2 = line2.Direction;
                            dir2.Normalize();
                            if(devDept.Geometry.Vector3D.AreOrthogonal(dir1, dir2))
                            {
                                var intersects = line1.IntersectWith(line2);
                                if(intersects.Count() == 1)
                                {
                                    if(line1.StartPoint.DistanceTo(intersects[0]) > 0.1 && line1.EndPoint.DistanceTo(intersects[0]) > 0.1)
                                    {
                                        var split1 = new devDept.Eyeshot.Entities.Line(line1.StartPoint, intersects[0])
                                        {
                                            LayerName = line1.LayerName,
                                            LineWeight = line1.LineWeight,
                                            LineTypeMethod = colorMethodType.byEntity
                                        };
                                        var split2 = new devDept.Eyeshot.Entities.Line(intersects[0], line1.EndPoint)
                                        {
                                            LayerName = line1.LayerName,
                                            LineWeight = line1.LineWeight,
                                            LineTypeMethod = colorMethodType.byEntity
                                        };
                                        this.designAutoCAD.Entities.Add(split1);
                                        this.designAutoCAD.Entities.Add(split2);
                                        this.designAutoCAD.Entities.Remove(line1);

                                        queue.Add(split1);
                                        queue.Add(split2);

                                        break;
                                    }

                                    if (line2.StartPoint.DistanceTo(intersects[0]) > 0.1 && line2.EndPoint.DistanceTo(intersects[0]) > 0.1)
                                    {
                                        var split1 = new devDept.Eyeshot.Entities.Line(line2.StartPoint, intersects[0])
                                        {
                                            LayerName = line2.LayerName,
                                            LineWeight = line2.LineWeight,
                                            LineTypeMethod = colorMethodType.byEntity
                                        };
                                        var split2 = new devDept.Eyeshot.Entities.Line(intersects[0], line2.EndPoint)
                                        {
                                            LayerName = line2.LayerName,
                                            LineWeight = line2.LineWeight,
                                            LineTypeMethod = colorMethodType.byEntity
                                        };
                                        this.designAutoCAD.Entities.Add(split1);
                                        this.designAutoCAD.Entities.Add(split2);
                                        this.designAutoCAD.Entities.Remove(line2);

                                        queue.Remove(line2);
                                        queue.Add(split1);
                                        queue.Add(split2);
                                    }
                                }
                            }
                        }
                    }
                    #endregion

                    ColorEntities(this.designAutoCAD, this.designAutoCAD.Entities, Verification.EqualColor);

                    // Sets the view as Top
                    this.designAutoCAD.SetView(viewType.Top);
                    this.designAutoCAD.ZoomFit();
                    this.designAutoCAD.Invalidate();
                }
            }

            /// AVEVA P&ID 파일을 화면에 표시한다.
            void ShowAVEVAPIDFile(string FilePath)
            {
                this.designAVEVA.Clear();
                if (System.IO.File.Exists(FilePath))
                {
                    devDept.Eyeshot.Translators.ReadAutodesk ra = new devDept.Eyeshot.Translators.ReadAutodesk(FilePath);
                    ra.DoWork();
                    ra.AddToScene(this.designAVEVA);

                    #region 멀티 라인들을 분할하여 추가한다.
                    var AddEntities = new List<Entity>();
                    this.designAVEVA.Entities.ForEach(x =>
                    {
                        if(x is Mesh mesh && (mesh.LayerName == "AS_PIPE" || mesh.LayerName == "AS_INST"))
                        {
                            int count = Convert.ToInt32(mesh.Vertices.Length * 0.5);
                            for (int i = 0;i < count - 1; ++i)
                            {
                                AddEntities.Add(new devDept.Eyeshot.Entities.Line(mesh.Vertices[i], mesh.Vertices[i + 1]) 
                                { 
                                    LayerName = mesh.LayerName,
                                    LineWeight = 3.0f,
                                    LineTypeMethod = colorMethodType.byEntity,
                                    Color = Verification.DiffColor,
                                    ColorMethod = colorMethodType.byEntity
                                });
                            }
                        }
                        else if (x is TabulatedSurface tf && (tf.LayerName == "AS_PIPE" || tf.LayerName == "AS_INST"))
                        {
                            int count = Convert.ToInt32(tf.ControlPoints.Length * 0.5);
                            for (int i = 0; i < count - 1; ++i)
                            {
                                AddEntities.Add(
                                    new devDept.Eyeshot.Entities.Line(
                                    new devDept.Geometry.Point3D(tf.ControlPoints[i, 0].X, tf.ControlPoints[i, 0].Y, 0),
                                    new devDept.Geometry.Point3D(tf.ControlPoints[i + 1, 0].X, tf.ControlPoints[i + 1, 0].Y, 0))
                                    {
                                        LayerName = tf.LayerName,
                                        LineWeight = 3.0f,
                                        LineTypeMethod = colorMethodType.byEntity,
                                        Color = Verification.DiffColor,
                                        ColorMethod = colorMethodType.byEntity
                                    }
                                );
                            }
                        }
                        else if (x is LinearPath lp)
                        {
                            int count = Convert.ToInt32(lp.Vertices.Length);
                            for (int i = 0; i < count - 1; ++i)
                            {
                                AddEntities.Add(new devDept.Eyeshot.Entities.Line(lp.Vertices[i], lp.Vertices[i + 1])
                                {
                                    LayerName = lp.LayerName,
                                    LineWeight = lp.LineWeight,
                                    LineTypeMethod = colorMethodType.byEntity
                                });
                            }
                        }
                    });
                    this.designAVEVA.Entities.RemoveAll(x => 
                    ((x is Mesh || x is TabulatedSurface) && (x.LayerName == "AS_PIPE" || x.LayerName == "AS_INST")) ||
                    (x is LinearPath));
                    this.designAVEVA.Entities.AddRange(AddEntities);
                    #endregion

                    ColorEntities(this.designAVEVA, this.designAVEVA.Entities, Verification.DiffColor);

                    #region 블럭 이름이 PSNODE, PENODE인 블럭은 보이지 않도록 한다.
                    this.designAVEVA.Entities.ForEach(x =>
                    {
                        if (x is BlockReference blkref && (blkref.BlockName == "PSNODE" || blkref.BlockName == "PENODE" || 
                        blkref.BlockName.StartsWith("ARROW")))
                            blkref.Visible = false;
                    });
                    #endregion

                    this.designAVEVA.SetView(viewType.Top);
                    this.designAVEVA.ZoomFit();
                    this.designAVEVA.Invalidate();
                }
            }

            if (this.radGridViewDocument.SelectedRows.Any() && this.radGridViewDocument.SelectedRows.First().DataBoundItem is Documents doc)
            {
                string dwgExtension = ".dwg";
                string ID2DrawingFolder = System.IO.Path.Combine(informations.FindID2LocalPath(doc.RefProjectCode), "drawings", "Native");
                string dwgFilePath = System.IO.Path.Combine(ID2DrawingFolder, $"{doc.DocumentNo}{dwgExtension}");
                this.radTextBoxID2DrawingFolder.Text = ID2DrawingFolder;
                ShowAutoCADFile(dwgFilePath);

                string AVEVAPIDFolder = this.radBrowseEditorAVEVAPIDFolder.Value;
                string AVEVAPIDFilePath = string.Empty;
                if (AVEVAPIDFolder != null)
                {
                    AVEVAPIDFilePath = System.IO.Path.Combine(AVEVAPIDFolder, $"{doc.DocumentNo}{dwgExtension}");
                    ShowAVEVAPIDFile(AVEVAPIDFilePath);
                }

                if(System.IO.File.Exists(dwgFilePath) && System.IO.File.Exists(AVEVAPIDFilePath))
                {
                    CompareAndMark(this.designAutoCAD, this.designAutoCAD.Entities, this.designAVEVA, this.designAVEVA.Entities);
                }

                if(OnDocumentSelected != null) OnDocumentSelected(doc);
            }
        }

        public void DocumentListBinding(List<ID2.Manager.Data.Models.Documents> docs)
        {
            this.radGridViewDocument.FilterDescriptors.Clear();
            this.radGridViewDocument.DataSource = new BindingList<ID2.Manager.Data.Models.Documents>(docs);
        }

        /// <summary>
        /// 주어진 두 엔터티 리스트를 비교하여 틀린 엔터티의 색상을 Yellow로 변경한다.
        /// </summary>
        /// <param name="entList1"></param>
        /// <param name="entList2"></param>
        private void CompareAndMark(Design design1, IList<Entity> entList1, Design design2, IList<Entity> entList2)
        {
            int[] equalEntitiesInV1 = new int[entList2.Count];
            bool[] equalEntitiesInV2 = new bool[entList2.Count];

            /// 서로 검사 가능한 타입인지 확인한다.
            bool CheckType(Entity ent1, Entity ent2)
            {
                return ent1.GetType() == ent2.GetType() ||
                    (ent1.GetType() == typeof(devDept.Eyeshot.Entities.Text) && ent2.GetType() == typeof(devDept.Eyeshot.Entities.MultilineText)) ||
                    (ent1.GetType() == typeof(devDept.Eyeshot.Entities.MultilineText) && ent2.GetType() == typeof(devDept.Eyeshot.Entities.Text)) ||
                    (ent1.GetType() == typeof(devDept.Eyeshot.Entities.Line) && ent2.GetType() == typeof(devDept.Eyeshot.Entities.TabulatedSurface));
            }

            try
            {
                for (int i = 0; i < entList1.Count(); i++)
                {
                    Entity entVp1 = entList1[i];
                    bool foundEqual = false;

                    for (int j = 0; j < entList2.Count(); j++)
                    {
                        Entity entVp2 = entList2[j];

                        if (!equalEntitiesInV2[j] && CheckType(entVp1, entVp2) && 
                            CompareIfEqual(this.designAutoCAD, entVp1, this.designAVEVA, entVp2))
                        {
                            equalEntitiesInV1[j] = i;
                            equalEntitiesInV2[j] = true;
                            foundEqual = true;
                            break;
                        }
                    }
                    if (!foundEqual)
                    {
                        ColorEntity(design1, entList1[i], Verification.DiffColor, colorMethodType.byEntity, false);
                    }
                }

                for (int j = 0; j < entList2.Count; j++)
                {
                    if (equalEntitiesInV2[j])
                    {
                        ColorEntity(design2, entList2[j], Verification.EqualColor, colorMethodType.byEntity, false);
                    }
                }
            }
            catch(Exception ex)
            {
                Console.Write($"Error : {ex.Message}");
            }
        }

        /// <summary>
        /// 주어진 두 엔터티를 비교한다.
        /// </summary>
        /// <param name="entVp1"></param>
        /// <param name="entVp2"></param>
        /// <returns></returns>
        private bool CompareIfEqual(Design design1, Entity entVp1, Design design2, Entity entVp2)
        {
            return AreEqual(design1, entVp1, design2, entVp2);
        }

        private bool CompareGEntityIfEqual(GEntity entVp1, GEntity entVp2)
        {
            return AreGEntityEqual(entVp1, entVp2);
        }

        /// <summary>
        /// 그래픽적으로 두 엔터티가 유사한지 검사한다.
        /// </summary>
        /// <param name="design1"></param>
        /// <param name="ent1"></param>
        /// <param name="design2"></param>
        /// <param name="ent2"></param>
        /// <returns></returns>
        private bool AreEqual(Design design1, Entity ent1, Design design2, Entity ent2)
        {
            if (ent1 is CompositeCurve cc1 && ent2 is CompositeCurve cc2)
            {
                if (cc1.CurveList.Count == cc2.CurveList.Count)
                {
                    int equalCurvesInListCount = 0;
                    foreach (var entC in cc1.CurveList)
                    {
                        foreach (var entC2 in cc2.CurveList)
                        {
                            if (entC.GetType() == entC2.GetType())
                            {
                                if (entC is Entity && entC2 is Entity && CompareIfEqual(design1, entC as Entity, design2, entC2 as Entity))
                                {
                                    equalCurvesInListCount++;
                                    break;
                                }
                                else if (entC is GEntity && entC2 is GEntity && CompareGEntityIfEqual(entC as GEntity, entC2 as GEntity))
                                {
                                    equalCurvesInListCount++;
                                    break;
                                }
                            }
                        }
                    }

                    if (cc1.CurveList.Count == equalCurvesInListCount)
                    {
                        return true;
                    }
                }
            }
            else if (ent1 is LinearPath lp1 && ent2 is LinearPath lp2)
            {
                if (lp1.Vertices.Length == lp2.Vertices.Length)
                {
                    for (int i = 0; i < lp1.Vertices.Length; i++)
                    {
                        if (lp1.Vertices[i].DistanceTo(lp2.Vertices[i]) > Verification.Tolerance)
                            return false;
                    }

                    return true;
                }
            }
            else if (ent1 is PlanarEntity pe1 && ent2 is PlanarEntity pe2)
            {
                if (ent1 is Arc arc1 && ent2 is Arc arc2)
                {
                    if (
                        arc1.Center.DistanceTo(arc2.Center) <= Verification.Tolerance &&
                        Math.Abs(arc1.Radius - arc2.Radius) <= Verification.Tolerance &&
                        Math.Abs(arc1.Domain.Min - arc2.Domain.Min) <= Verification.Tolerance &&
                        Math.Abs(arc1.Domain.Max - arc2.Domain.Max) <= Verification.Tolerance
                        )
                    {
                        return true;
                    }
                }
                else if (ent1 is Circle c1 && ent2 is Circle c2)
                {
                    if (
                        c1.Center.DistanceTo(c2.Center) <= Verification.Tolerance &&
                        Math.Abs(c1.Radius - c2.Radius) <= Verification.Tolerance
                        )
                    {
                        return true;
                    }
                }
                else if (ent1 is EllipticalArc e1 && ent2 is EllipticalArc e2)
                {
                    if (
                        e1.Center.DistanceTo(e2.Center) <= Verification.Tolerance &&
                        Math.Abs(e1.RadiusX - e2.RadiusX) <= Verification.Tolerance &&
                        Math.Abs(e1.RadiusY - e2.RadiusY) <= Verification.Tolerance &&
                        Math.Abs(e1.Domain.Low - e2.Domain.Low) <= Verification.Tolerance &&
                        Math.Abs(e1.Domain.High - e2.Domain.High) <= Verification.Tolerance
                    )
                    {
                        return true;
                    }
                }
                else if (ent1 is Ellipse el1 && ent2 is Ellipse el2)
                {
                    if (
                        el1.Center.DistanceTo(el2.Center) <= Verification.Tolerance &&
                        Math.Abs(el1.RadiusX - el2.RadiusX) <= Verification.Tolerance &&
                        Math.Abs(el1.RadiusY - el2.RadiusY) <= Verification.Tolerance
                    )
                    {
                        return true;
                    }
                }

                else if (ent1 is Text)
                {
                    if (ent1 is Dimension dim1 && ent2 is Dimension dim2)
                    {
                        if (
                            dim1.InsertionPoint.DistanceTo(dim2.InsertionPoint) <= Verification.Tolerance &&
                            dim1.DimLinePosition.DistanceTo(dim2.DimLinePosition) <= Verification.Tolerance
                            )
                        {
                            if (ent1 is AngularDim ad1 && ent2 is AngularDim ad2)
                            {
                                if (
                                    ad1.ExtLine1.DistanceTo(ad2.ExtLine1) <= Verification.Tolerance &&
                                    ad1.ExtLine2.DistanceTo(ad2.ExtLine2) <= Verification.Tolerance &&
                                    Math.Abs(ad1.StartAngle - ad2.StartAngle) <= Verification.Tolerance &&
                                    Math.Abs(ad1.EndAngle - ad2.EndAngle) <= Verification.Tolerance &&
                                    Math.Abs(ad1.Radius - ad2.Radius) <= Verification.Tolerance
                                    )
                                {
                                    return true;
                                }
                            }
                            else if (ent1 is LinearDim ld1 && ent2 is LinearDim ld2)
                            {
                                if (
                                    ld1.ExtLine1.DistanceTo(ld2.ExtLine1) <= Verification.Tolerance &&
                                    ld1.ExtLine2.DistanceTo(ld2.ExtLine2) <= Verification.Tolerance
                                    )
                                {
                                    return true;
                                }
                            }
                            else if (ent1 is DiametricDim dd1 && ent2 is DiametricDim dd2)
                            {
                                if (
                                    Math.Abs(dd1.Distance - dd2.Distance) <= Verification.Tolerance &&
                                    Math.Abs(dd1.Radius - dd2.Radius) <= Verification.Tolerance &&
                                    Math.Abs(dd1.CenterMarkSize - dd2.CenterMarkSize) <= Verification.Tolerance
                                )
                                {
                                    return true;
                                }
                            }
                            else if (ent1 is RadialDim rd1 && ent2 is RadialDim rd2)
                            {
                                if (
                                    Math.Abs(rd1.Radius - rd2.Radius) <= Verification.Tolerance &&
                                    Math.Abs(rd1.CenterMarkSize - rd2.CenterMarkSize) <= Verification.Tolerance
                                )
                                {
                                    return true;
                                }
                            }
                            else if (ent1 is OrdinateDim od1 && ent2 is OrdinateDim od2)
                            {
                                if (
                                    od1.DefiningPoint.DistanceTo(od2.DefiningPoint) <= Verification.Tolerance &&
                                    od1.Origin.DistanceTo(od2.Origin) <= Verification.Tolerance &&
                                    od1.LeaderEndPoint.DistanceTo(od2.LeaderEndPoint) <= Verification.Tolerance
                                )
                                {
                                    return true;
                                }
                            }
                            else
                            {
                                Console.Write("Type " + ent1.GetType() + " not implemented.");
                                return true;
                            }
                        }
                    }

                    else if (ent1 is devDept.Eyeshot.Entities.Attribute att1 && ent2 is devDept.Eyeshot.Entities.Attribute att2)
                    {
                        if (
                            att1.Value == att2.Value &&
                            att1.InsertionPoint.DistanceTo(att2.InsertionPoint) <= Verification.Tolerance
                            )
                        {
                            return true;
                        }
                    }
                    else
                    {
                        Text tx1 = (Text)ent1;
                        Text tx2 = (Text)ent2;

                        if (
                            tx1.InsertionPoint.DistanceTo(tx2.InsertionPoint) <= Verification.Tolerance &&
                            tx1.TextString.Trim() == tx2.TextString.Trim() &&
                            Math.Abs(tx1.WidthFactor - tx2.WidthFactor) <= Verification.Tolerance &&
                            Math.Abs(tx1.Height - tx2.Height) <= Verification.Tolerance
                            )
                        {
                            return true;
                        }
                    }
                }
            }
            else if (ent1 is Line line1 && ent2 is Line line2)
            {
                var dir1 = line1.Direction;
                dir1.Normalize();
                var dir2 = line2.Direction;
                dir2.Normalize();
                if (devDept.Geometry.Vector3D.AreParallel(dir1, dir2, 0.1) &&
                    Math.Abs(line1.Length() - line2.Length()) <= Verification.Tolerance &&
                    line1.MidPoint.DistanceTo(line2.MidPoint) <= Verification.Tolerance
                )
                {
                    return true;
                }
            }
            else if (ent1 is Line && ent2 is devDept.Eyeshot.Entities.TabulatedSurface lwpolyline && lwpolyline.ControlPoints.Length == 4)
            {
                line1 = ent1 as Line;
                var start = new devDept.Geometry.Point3D(lwpolyline.ControlPoints[0, 0].X, lwpolyline.ControlPoints[0, 0].Y, 0);
                var end = new devDept.Geometry.Point3D(lwpolyline.ControlPoints[1, 0].X, lwpolyline.ControlPoints[1, 0].Y, 0);
                var vec = new devDept.Geometry.Vector3D(start, end);
                vec.Normalize();
                var dir = line1.Direction.Clone() as devDept.Geometry.Vector3D;
                dir.Normalize();

                if (
                    devDept.Geometry.Vector3D.AreParallel(dir, vec) &&
                    line1.StartPoint.DistanceTo(start) <= Verification.Tolerance &&
                    line1.EndPoint.DistanceTo(end) <= Verification.Tolerance
                )
                {
                    return true;
                }
            }
            else if (ent1 is devDept.Eyeshot.Entities.Point point1 && ent2 is devDept.Eyeshot.Entities.Point point2)
            {
                if (point1.Position.DistanceTo(point2.Position) <= Verification.Tolerance)
                {
                    return true;
                }
            }
            else if (ent1 is Curve cu1 && ent2 is Curve cu2)
            {
                if (
                    cu1.ControlPoints.Length == cu2.ControlPoints.Length &&
                    cu1.KnotVector.Length == cu2.KnotVector.Length &&
                    cu1.Degree == cu2.Degree
                    )
                {
                    for (int k = 0; k < cu1.ControlPoints.Length; k++)
                    {
                        if (cu1.ControlPoints[k].DistanceTo(cu2.ControlPoints[k]) > Verification.Tolerance)
                        {
                            return false;
                        }
                    }

                    for (int k = 0; k < cu1.KnotVector.Length; k++)
                    {
                        if (cu1.KnotVector[k] != cu2.KnotVector[k])
                        {
                            return false;
                        }
                    }

                    return true;
                }
            }
            else if (ent1 is Mesh m1 && ent2 is Mesh m2 && m1.Vertices.Count() == m2.Vertices.Count())
            {
                for (int i = 0; i < m1.Vertices.Count(); ++i)
                {
                    if (m1.Vertices[i].DistanceTo(m2.Vertices[i]) > Verification.Tolerance) return false;
                }

                return true;
            }
            else if (ent1 is BlockReference blkref1 && ent2 is BlockReference blkref2)
            {
                int equalCurvesInEntityList = 0;

                #region Point 제거 및 LinePath를 라인으로 분리
                var entities1 = blkref1.Explode(design1.Blocks).Where(x => x.LayerName != "AS_PORT" && !(x is devDept.Eyeshot.Entities.Point)).ToList();
                var coll1 = new List<Entity>();
                entities1.ForEach(x =>
                {
                    if (x is LinearPath lp)
                    {
                        for (int i = 0; i < lp.Vertices.Length - 1; ++i)
                        {
                            if (lp.Vertices[i].DistanceTo(lp.Vertices[i + 1]) < 0.1) continue;
                            coll1.Add(new Line(lp.Vertices[i], lp.Vertices[i + 1]));
                        }
                    }
                    else
                    {
                        coll1.Add(x);
                    }
                });
                #endregion

                #region Point 및 Nesting Block 제거 및 LinePath를 라인으로 분리
                var entities2 = blkref2.Explode(design2.Blocks).Where(x => 
                !(x is devDept.Eyeshot.Entities.BlockReference blkref && blkref.BlockName == "PORT") && !(x is devDept.Eyeshot.Entities.Point)).ToList();
                var coll2 = new List<Entity>();
                entities2.ForEach(x =>
                {
                    if (x is LinearPath lp)
                    {
                        for (int i = 0; i < lp.Vertices.Length - 1; ++i)
                        {
                            if (lp.Vertices[i].DistanceTo(lp.Vertices[i + 1]) < 0.1) continue;
                            coll2.Add(new Line(lp.Vertices[i], lp.Vertices[i + 1]));
                        }
                    }
                    else if (x is devDept.Eyeshot.Entities.Attribute attr)
                    {
                        if (!attr.Invisible) coll2.Add(attr);
                    }
                    else if (x.GetType().Name == "AttributeReferenceData")
                    {
                    }
                    else
                    {
                        coll2.Add(x);
                    }
                });
                #endregion

                if (coll1.Count != coll2.Count) return false;

                foreach (var entC in coll1)
                {
                    foreach (var entC2 in coll2)
                    {
                        if (entC.GetType() == entC2.GetType())
                        {
                            if (entC is Entity && entC2 is Entity && CompareIfEqual(design1, entC as Entity, design2, entC2 as Entity))
                            {
                                equalCurvesInEntityList++;
                                break;
                            }
                        }
                    }
                }

                if (coll1.Count == equalCurvesInEntityList)
                {
                    return true;
                }
            }
            else
            {
                Console.Write("Type " + ent1.GetType() + " not implemented.");
                return false;
            }

            return false;
        }

        private bool AreGEntityEqual(GEntity ent1, GEntity ent2)
        {
            if (ent1 is GCompositeCurve cc1 && ent2 is GCompositeCurve cc2)
            {
                if (cc1.CurveList.Count == cc2.CurveList.Count)
                {
                    int equalCurvesInListCount = 0;
                    foreach (var entC in cc1.CurveList)
                    {
                        foreach (var entC2 in cc2.CurveList)
                        {
                            if (entC.GetType() == entC2.GetType())
                            {
                                if (entC is GEntity && entC2 is GEntity && CompareGEntityIfEqual(entC as GEntity, entC2 as GEntity))
                                {
                                    equalCurvesInListCount++;
                                    break;
                                }
                            }
                        }
                    }

                    if (cc1.CurveList.Count == equalCurvesInListCount)
                    {
                        return true;
                    }
                }
            }
            else if (ent1 is GLinearPath lp1 && ent2 is GLinearPath lp2)
            {
                if (lp1.Vertices.Length == lp2.Vertices.Length)
                {
                    for (int i = 0; i < lp1.Vertices.Length; i++)
                    {
                        if (lp1.Vertices[i].DistanceTo(lp2.Vertices[i]) > Verification.Tolerance)
                            return false;
                    }
                    return true;
                }
            }

            else if (ent1 is GPlanarEntity pe1 && ent2 is GPlanarEntity pe2)
            {
                if (
                    pe1.Plane.AxisZ == pe2.Plane.AxisZ &&
                    pe1.Plane.AxisX == pe2.Plane.AxisX
                    )
                {
                    if (ent1 is GArc arc1 && ent2 is GArc arc2)
                    {
                        if (
                            arc1.Center.DistanceTo(arc2.Center) <= Verification.Tolerance &&
                            Math.Abs(arc1.Radius - arc2.Radius) <= Verification.Tolerance &&
                            Math.Abs(arc1.Domain.Min - arc2.Domain.Min) <= Verification.Tolerance &&
                            Math.Abs(arc1.Domain.Max - arc2.Domain.Max) <= Verification.Tolerance
                            )
                        {
                            return true;
                        }
                    }
                    else if (ent1 is GCircle c1 && ent2 is GCircle c2)
                    {
                        if (c1.Center.DistanceTo(c2.Center) <= Verification.Tolerance && 
                            Math.Abs(c1.Radius - c2.Radius) <= Verification.Tolerance)
                        {
                            return true;
                        }
                    }
                    else if (ent1 is GEllipticalArc e1 && ent2 is GEllipticalArc e2)
                    {
                        if (
                            e1.Center.DistanceTo(e2.Center) <= Verification.Tolerance &&
                            Math.Abs(e1.RadiusX - e2.RadiusX) <= Verification.Tolerance &&
                            Math.Abs(e1.RadiusY - e2.RadiusY) <= Verification.Tolerance &&
                            Math.Abs(e1.Domain.Low - e2.Domain.Low) <= Verification.Tolerance &&
                            Math.Abs(e1.Domain.High - e2.Domain.High) <= Verification.Tolerance
                        )
                        {
                            return true;
                        }
                    }
                    else if (ent1 is GEllipse el1 && ent2 is GEllipse el2)
                    {
                        if (
                            el1.Center.DistanceTo(el2.Center) <= Verification.Tolerance &&
                            Math.Abs(el1.RadiusX - el2.RadiusX) <= Verification.Tolerance &&
                            Math.Abs(el1.RadiusY - el2.RadiusY) <= Verification.Tolerance
                        )
                        {
                            return true;
                        }
                    }
                    else
                    {
                        Console.Write("Type " + ent1.GetType() + " not implemented.");
                        return true;
                    }
                }
            }

            else if (ent1 is GLine line1 && ent2 is GLine line2)
            {
                if (line1.StartPoint.DistanceTo(line2.StartPoint) <= Verification.Tolerance && 
                    line1.EndPoint.DistanceTo(line2.EndPoint) <= Verification.Tolerance
                )
                {
                    return true;
                }
            }
#if NURBS
            else if (ent1 is Curve)
            {
                Curve cu1 = (Curve)ent1;
                Curve cu2 = (Curve)ent2;

                if (
                    cu1.ControlPoints.Length == cu2.ControlPoints.Length &&
                    cu1.KnotVector.Length == cu2.KnotVector.Length &&
                    cu1.Degree == cu2.Degree
                    )
                {
                    for (int k = 0; k < cu1.ControlPoints.Length; k++)
                    {
                        if (cu1.ControlPoints[k] != cu2.ControlPoints[k])
                        {
                            return false;
                        }
                    }

                    for (int k = 0; k < cu1.KnotVector.Length; k++)
                    {
                        if (cu1.KnotVector[k] != cu2.KnotVector[k])
                        {
                            return false;
                        }
                    }

                    return true;
                }
            }
#endif

            else
            {
                Console.Write("Type " + ent1.GetType() + " not implemented.");
                return true;
            }
            return false;
        }

        #region Camera Sync
        private void CameraChanged(object sender, devDept.Eyeshot.Workspace.CameraMoveEventArgs e)
        {
            if (sender == this.designAutoCAD)
                SyncCamera(this.designAutoCAD, this.designAVEVA);
            else
                SyncCamera(this.designAVEVA, this.designAutoCAD);
        }

        private void SyncCamera(Design designMovedCamera, Design designCameraToMove)
        {
            Camera savedCamera;
            designMovedCamera.SaveView(out savedCamera);

            // restores the camera to the other model
            designCameraToMove.RestoreView(savedCamera);
            designCameraToMove.AdjustNearAndFarPlanes();
            designCameraToMove.Invalidate();
        }
        #endregion
    }
}
